@@ -7,7 +7,8 @@ errors = {
     'directory_already_exists': 205,
     'validation_failed': 206,
     'failed_to_wheel': 207,
-    'could_not_install_module': 208,
+    'failed_to_retrieve_index_info': 208,
     'failed_to_create_virtualenv': 209,
-    'cannot_exclude_main_module': 210
+    'cannot_exclude_main_package': 210,
+    'failed_to_install_package': 211
 }
