@@ -1,18 +1,32 @@
-import logger
-import logging
+########
+# Copyright (c) 2014 GigaSpaces Technologies Ltd. All rights reserved
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#        http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+#    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#    * See the License for the specific language governing permissions and
+#    * limitations under the License.
+
 import os
 import sys
-import shutil
-import click
-import uuid
-import tempfile
 import json
+import shutil
+import logging
+import tempfile
+
+import click
+
+from . import logger, utils, codes
 
-import utils
-import codes
 
 REQUIREMENT_FILE_NAMES = ['dev-requirements.txt', 'requirements.txt']
-METADATA_FILE_NAME = 'module.json'
+METADATA_FILE_NAME = 'package.json'
 DEFAULT_WHEELS_PATH = 'wheels'
 
 lgr = logger.init()
@@ -27,37 +41,36 @@ class Wagon():
     - validate
     - get_metadata_from_archive
     """
+
     def __init__(self, source, verbose=False):
         """Source depends on the context in which
         the class is instantiated.
 
         When using `create`, source can be a path to a local setup.py
-        containing directory; a URL to a GitHub like module archive or a
-        name of a PyPI module in the format: MODULE_NAME==MODULE_VERSION.
+        containing directory; a URL to a GitHub like package archive, a
+        name of a PyPI package in the format: PACKAGE_NAME==PACKAGE_VERSION
+        or a versionless PyPI PACKAGE_NAME.
 
         When using `install` or `validate`, source can be either a path
         to a local or a URL based Wagon archived tar.gz file.
         """
-        if verbose:
-            lgr.setLevel(logging.DEBUG)
-        else:
-            lgr.setLevel(logging.INFO)
+        lgr.setLevel(logging.DEBUG if verbose else logging.INFO)
         self.source = source
 
-    def create(self, with_requirements=None, force=False,
-               keep_wheels=False, excluded_modules=None,
+    def create(self, with_requirements='', force=False,
+               keep_wheels=False, excluded_packages=None,
                archive_destination_dir='.', python_versions=None,
-               validate=False):
+               validate=False, wheel_args='', format='tar.gz'):
         """Creates a Wagon archive and returns its path.
 
         This currently only creates tar.gz archives. The `install`
         method assumes tar.gz when installing on Windows as well.
 
-        Module name and version are extracted from the setup.py file
-        of the `source` or from the MODULE_NAME==MODULE_VERSION if the source
-        is a PyPI module.
+        Package name and version are extracted from the setup.py file
+        of the `source` or from the PACKAGE_NAME==PACKAGE_VERSION if the source
+        is a PyPI package.
 
-        Excluded modules will be removed from the archive even if they are
+        Excluded packages will be removed from the archive even if they are
         required for installation and their wheel names will be appended
         to the metadata for later analysis/validation.
 
@@ -70,46 +83,51 @@ class Wagon():
         will be automatically extracted from either the GitHub archive URL
         or the local path provided provided in `source`.
         """
-        lgr.info('Creating module package for {0}...'.format(self.source))
+        lgr.info('Creating archive for {0}...'.format(self.source))
         source = self.get_source(self.source)
-        module_name, module_version = \
+        package_name, package_version = \
             self.get_source_name_and_version(source)
 
-        excluded_modules = excluded_modules or []
-        if excluded_modules:
-            lgr.warn('Note that excluding modules may make the archive '
+        excluded_packages = excluded_packages or []
+        if excluded_packages:
+            lgr.warn('Note that excluding packages may make the archive '
                      'non-installable.')
-        if module_name in excluded_modules:
-            lgr.error('You cannot exclude the module you are trying to wheel.')
-            sys.exit(codes.errors['cannot_exclude_main_module'])
+        if package_name in excluded_packages:
+            lgr.error('You cannot exclude the package '
+                      'you are trying to wheel.')
+            sys.exit(codes.errors['cannot_exclude_main_package'])
 
-        wheels_path = os.path.join(module_name, DEFAULT_WHEELS_PATH)
-        self.handle_output_directory(module_name, force)
+        tempdir = tempfile.mkdtemp()
+        self.workdir = os.path.join(tempdir, package_name)
+        wheels_path = os.path.join(self.workdir, DEFAULT_WHEELS_PATH)
 
-        if with_requirements == '.':
+        if with_requirements:
             with_requirements = self._get_default_requirement_files(source)
-        elif with_requirements:
-            with_requirements = [with_requirements]
 
-        wheels, excluded_wheels = utils.wheel(
-            source, with_requirements, wheels_path, excluded_modules)
+        try:
+            wheels, excluded_wheels = utils.wheel(
+                source, with_requirements, wheels_path, excluded_packages,
+                wheel_args)
+        finally:
+            if self.remove_source_after_process:
+                shutil.rmtree(source, ignore_errors=True)
+
         self.platform = utils.get_platform_for_set_of_wheels(wheels_path)
-        if python_versions:
-            self.python_versions = ['py{0}'.format(v) for v in python_versions]
-        else:
-            self.python_versions = [utils.get_python_version()]
+        lgr.debug('Platform is: {0}'.format(self.platform))
+        self.python_versions = self._set_python_versions(python_versions)
 
-        archive_file = self.set_archive_name(module_name, module_version)
+        if not os.path.isdir(archive_destination_dir):
+            os.makedirs(archive_destination_dir)
+        archive_file = self._set_archive_name(package_name, package_version)
         archive_path = os.path.join(archive_destination_dir, archive_file)
 
-        self.handle_output_file(archive_path, force)
-        self.generate_metadata_file(wheels, excluded_wheels)
-
-        utils.tar(module_name, archive_path)
+        self._handle_output_file(archive_path, force)
+        self._generate_metadata_file(wheels, excluded_wheels)
 
+        self._archive(format, self.workdir, archive_path)
         if not keep_wheels:
             lgr.debug('Cleaning up...')
-            shutil.rmtree(module_name)
+            shutil.rmtree(self.workdir, ignore_errors=True)
 
         if validate:
             self.source = archive_path
@@ -117,8 +135,24 @@ class Wagon():
         lgr.info('Process complete!')
         return archive_path
 
-    def install(self, virtualenv=None, requirements_file=None, upgrade=False,
-                ignore_platform=False):
+    @staticmethod
+    def _set_python_versions(python_versions):
+        if python_versions:
+            return ['py{0}'.format(v) for v in python_versions]
+        else:
+            return [utils.get_python_version()]
+
+    def _archive(self, format, source_path, archive_path):
+        if format == 'tar.gz':
+            utils.tar(source_path, archive_path)
+        elif format == 'zip':
+            utils.zip(self.workdir, archive_path)
+        else:
+            sys.exit('Unsupported archive format to create '
+                     '(Must be one of [zip, tar.gz]).')
+
+    def install(self, virtualenv='', requirements_file='', upgrade=False,
+                ignore_platform=False, install_args=''):
         """Installs a Wagon archive.
 
         This can install in a provided `virtualenv` or in the current
@@ -133,21 +167,22 @@ class Wagon():
         """
         lgr.info('Installing {0}'.format(self.source))
         source = self.get_source(self.source)
-        with open(os.path.join(source, 'module.json'), 'r') as f:
+        with open(os.path.join(source, METADATA_FILE_NAME), 'r') as f:
             metadata = json.loads(f.read())
         supported_platform = metadata['supported_platform']
         if not ignore_platform and supported_platform != 'any':
             lgr.debug('Validating Platform {0} is supported...'.format(
                 supported_platform))
-            machine_platform = utils.get_machine_platform()
+            machine_platform = utils.get_platform()
             if machine_platform != supported_platform:
-                lgr.error('Platform unsupported for module ({0}).'.format(
+                lgr.error('Platform unsupported for package ({0}).'.format(
                     machine_platform))
-                sys.exit(codes.errors['unsupported_platform_for_module'])
+                sys.exit(codes.errors['unsupported_platform_for_package'])
 
         wheels_path = os.path.join(source, DEFAULT_WHEELS_PATH)
-        utils.install_module(metadata['module_name'], wheels_path, virtualenv,
-                             requirements_file, upgrade)
+        utils.install_package(
+            metadata['package_name'], wheels_path, virtualenv,
+            requirements_file, upgrade, install_args)
 
     def validate(self):
         """Validates a Wagon archive. Return True if succeeds, False otherwise.
@@ -155,7 +190,7 @@ class Wagon():
 
         This will test that some of the metadata is solid, that
         the required wheels are present within the archives and that
-        the module is installable.
+        the package is installable.
 
         Note that if the metadata file is corrupted, validation
         of the required wheels will be corrupted as well, since validation
@@ -164,7 +199,7 @@ class Wagon():
         """
         lgr.info('Validating {0}'.format(self.source))
         source = self.get_source(self.source)
-        with open(os.path.join(source, 'module.json'), 'r') as f:
+        with open(os.path.join(source, METADATA_FILE_NAME), 'r') as f:
             metadata = json.loads(f.read())
         wheels_path = os.path.join(source, DEFAULT_WHEELS_PATH)
         validation_errors = []
@@ -175,10 +210,10 @@ class Wagon():
                 '`supported_platform` key not found in metadata file '
                 'and is required for installation.')
 
-        lgr.debug('Verifying that `module_name` key is in metadata...')
-        if not metadata.get('module_name'):
+        lgr.debug('Verifying that `package_name` key is in metadata...')
+        if not metadata.get('package_name'):
             validation_errors.append(
-                '`module_name` key not found in metadata file '
+                '`package_name` key not found in metadata file '
                 'and is required for installation.')
 
         lgr.debug('Verifying that `wheels` key is in metadata...')
@@ -194,7 +229,7 @@ class Wagon():
             if not os.path.isfile(os.path.join(wheels_path, wheel)):
                 validation_errors.append('Missing wheel: {0}'.format(wheel))
 
-        lgr.debug('Testing module installation...')
+        lgr.debug('Testing package installation...')
         excluded_wheels = metadata.get('excluded_wheels')
         if excluded_wheels:
             for wheel in excluded_wheels:
@@ -228,7 +263,7 @@ class Wagon():
         """
         lgr.debug('Retrieving Metadata for: {0}'.format(self.source))
         source = self.get_source(self.source)
-        with open(os.path.join(source, 'module.json'), 'r') as f:
+        with open(os.path.join(source, METADATA_FILE_NAME), 'r') as f:
             metadata = json.loads(f.read())
         shutil.rmtree(source)
         return metadata
@@ -239,11 +274,12 @@ class Wagon():
             return [os.path.join(source, f) for f in REQUIREMENT_FILE_NAMES
                     if os.path.isfile(os.path.join(source, f))]
 
-    def generate_metadata_file(self, wheels, excluded_wheels):
-        """This generates a metadata file for the module.
+    def _generate_metadata_file(self, wheels, excluded_wheels):
+        """This generates a metadata file for the package.
         """
         lgr.debug('Generating Metadata...')
         metadata = {
+            'created_by_wagon_version': utils.get_wagon_version(),
             'archive_name': self.archive,
             'supported_platform': self.platform,
             'supported_python_versions': self.python_versions,
@@ -252,9 +288,9 @@ class Wagon():
                 'distribution_version': None,
                 'distribution_release': None,
             },
-            'module_name': self.name,
-            'module_source': self.source,
-            'module_version': self.version,
+            'package_name': self.name,
+            'package_version': self.version,
+            'package_source': self.source,
             'wheels': wheels,
             'excluded_wheels': excluded_wheels
         }
@@ -269,24 +305,24 @@ class Wagon():
 
         formatted_metadata = json.dumps(metadata, indent=4, sort_keys=True)
         lgr.debug('Metadata is: {0}'.format(formatted_metadata))
-        output_path = os.path.join(self.name, METADATA_FILE_NAME)
+        output_path = os.path.join(self.workdir, METADATA_FILE_NAME)
         with open(output_path, 'w') as f:
             lgr.debug('Writing metadata to file: {0}'.format(output_path))
             f.write(formatted_metadata)
 
-    def set_archive_name(self, module_name, module_version):
+    def _set_archive_name(self, package_name, package_version):
         """Sets the format of the output archive file.
 
         We should aspire for the name of the archive to be
         as compatible as possible with the wheel naming convention
         described here:
         https://www.python.org/dev/peps/pep-0491/#file-name-convention,
-        as we've basically providing a "wheel" of our module.
+        as we've basically providing a "wheel" of our package.
         """
-        module_name = module_name.replace('-', '_')
+        package_name = package_name.replace('-', '_')
         python_versions = '.'.join(self.python_versions)
 
-        archive = [module_name, module_version, python_versions, 'none',
+        archive = [package_name, package_version, python_versions, 'none',
                    self.platform, 'none', 'none']
 
         if utils.IS_LINUX and self.platform != 'any':
@@ -296,19 +332,34 @@ class Wagon():
             if release:
                 archive[6] = release
 
-        self.archive = '{0}.tar.gz'.format('-'.join(archive))
+        self.archive = '{0}.wgn'.format('-'.join(archive))
         return self.archive
 
     def get_source(self, source):
         """If necessary, downloads and extracts the source.
 
-        If the source is a url to a module's tar file,
+        If the source is a url to a package's tar file,
         this will download the source and extract it to a temporary directory.
+
+        If the source is neither a url nor a local path, and is not provided
+        as PACKAGE_NAME==PACKAGE_VERSION, the provided source string
+        will be regarded as the source, which, by default, will assume
+        that the string is a name of a package in PyPI.
         """
         def extract_source(source, destination):
-            utils.untar(source, destination)
-            return os.path.join(
+            try:
+                utils.untar(source, destination)
+            except:
+                utils.unzip(source, destination)
+            source = os.path.join(
                 destination, [d for d in os.walk(destination).next()[1]][0])
+            if not os.path.join(source, 'setup.py'):
+                sys.exit('Source does not seem to be a Python package. '
+                         'A source archive must contain a single parent '
+                         'directory containing a setup.py file.')
+            return source
+
+        self.remove_source_after_process = False
 
         lgr.debug('Retrieving source...')
         if '://' in source:
@@ -316,8 +367,10 @@ class Wagon():
             schema = split[0]
             if schema in ['http', 'https']:
                 tmpdir = tempfile.mkdtemp()
-                tmpfile = os.path.join(tmpdir, str(uuid.uuid4()))
+                fd, tmpfile = tempfile.mkstemp()
+                os.close(fd)
                 try:
+                    self.remove_source_after_process = True
                     utils.download_file(source, tmpfile)
                     source = extract_source(tmpfile, tmpdir)
                 finally:
@@ -326,46 +379,46 @@ class Wagon():
                 lgr.error('Source URL type {0} is not supported'.format(
                     schema))
                 sys.exit(codes.errors['unsupported_url_type'])
-        elif os.path.isfile(source) and source.endswith('.tar.gz'):
+        elif os.path.isfile(source):
+            self.remove_source_after_process = True
             tmpdir = tempfile.mkdtemp()
             source = extract_source(source, tmpdir)
         elif os.path.isdir(source):
             source = os.path.expanduser(source)
-        elif '==' in source:
-            pass
-        else:
-            lgr.error('Path to source {0} does not exist.'.format(source))
-            sys.exit(codes.errors['nonexistent_source_path'])
         lgr.debug('Source is: {0}'.format(source))
         return source
 
     def get_source_name_and_version(self, source):
-        """Retrieves the source module's name and version.
+        """Retrieves the source package's name and version.
 
         If the source is a path, the name and version will be retrieved
         by querying the setup.py file in the path.
 
-        If the source is of format MODULE==VERSION, they will be used as
+        If the source is PACKAGE_NAME==PACKAGE_VERSION, they will be used as
         the name and version.
+
+        If the source is PACKAGE_NAME, the version will be extracted from
+        the wheel of the latest version.
         """
         if os.path.isfile(os.path.join(source, 'setup.py')):
             lgr.debug('setup.py file found. Retrieving name and version...')
             setuppy_path = os.path.join(source, 'setup.py')
-            self.name = utils.run('python {0} --name'.format(
-                setuppy_path)).aggr_stdout.strip('\n')
-            self.version = utils.run('python {0} --version'.format(
-                setuppy_path)).aggr_stdout.strip('\n')
+            self.name = utils.run('{0} {1} --name'.format(
+                sys.executable, setuppy_path)).aggr_stdout.rstrip('\r\n')
+            self.version = utils.run('{0} {1} --version'.format(
+                sys.executable, setuppy_path)).aggr_stdout.rstrip('\r\n')
         # TODO: maybe we don't want to be that explicit and allow using >=
-        # TODO: or just a module name...
         elif '==' in source:
-            lgr.debug('Retrieving name and version...')
             self.name, self.version = source.split('==')
-        lgr.info('Module name: {0}'.format(self.name))
-        lgr.info('Module version: {0}'.format(self.version))
+        else:
+            self.name = source
+            self.version = utils.get_package_version_from_wheel_name(source)
+        lgr.info('Package name: {0}'.format(self.name))
+        lgr.info('Package version: {0}'.format(self.version))
         return self.name, self.version
 
-    def handle_output_file(self, archive, force):
-        """Handles the output tar.
+    def _handle_output_file(self, archive, force):
+        """Handles the output file.
 
         removes the output file if required, else, notifies
         that it already exists.
@@ -378,15 +431,6 @@ class Wagon():
                       'the -f flag to overwrite.'.format(archive))
             sys.exit(codes.errors['archive_already_exists'])
 
-    def handle_output_directory(self, wheels_path, force):
-        if os.path.isdir(wheels_path):
-            if force:
-                shutil.rmtree(wheels_path)
-            else:
-                lgr.error('Directory {0} already exists. Please remove it and '
-                          'run this again.'.format(wheels_path))
-                sys.exit(codes.errors['directory_already_exists'])
-
 
 @click.group()
 def main():
@@ -395,15 +439,18 @@ def main():
 
 @click.command()
 @click.option('-s', '--source', required=True,
-              help='Source URL, Path or Module name.')
-@click.option('-r', '--with-requirements', required=False,
+              help='Source URL, Path or Package name.')
+@click.option('-r', '--with-requirements', required=False, is_flag=True,
               help='Whether to also pack wheels from a requirements file.')
+@click.option('-t', '--format', required=False, default='tar.gz',
+              type=click.Choice(['tar.gz', 'zip']),
+              help='Which file format to generate.')
 @click.option('-f', '--force', default=False, is_flag=True,
               help='Force overwriting existing output file.')
 @click.option('--keep-wheels', default=False, is_flag=True,
               help='Keep wheels path after creation.')
 @click.option('-x', '--exclude', default=None, multiple=True,
-              help='Specific modules to exclude from the archive. '
+              help='Specific packages to exclude from the archive. '
                    'This argument can be provided multiple times.')
 @click.option('-o', '--output-directory', default='.',
               help='Output directory for the archive.')
@@ -412,10 +459,13 @@ def main():
                    'This argument can be provided multiple times.')
 @click.option('--validate', default=False, is_flag=True,
               help='Runs a postcreation validation on the archive.')
+@click.option('-a', '--wheel-args', required=False,
+              help='Allows to pass additional arguments to `pip wheel`. '
+                   '(e.g. --no-cache-dir -c constains.txt')
 @click.option('-v', '--verbose', default=False, is_flag=True)
-def create(source, with_requirements, force, keep_wheels, exclude,
-           output_directory, pyver, validate, verbose):
-    """Creates a Python module's wheel base archive.
+def create(source, with_requirements, format, force, keep_wheels, exclude,
+           output_directory, pyver, validate, wheel_args, verbose):
+    r"""Creates a Python package's wheel base archive.
 
     \b
     Example sources:
@@ -423,40 +473,50 @@ def create(source, with_requirements, force, keep_wheels, exclude,
     master.tar.gz
     - ~/repos/cloudify-script-plugin
     - cloudify-script-plugin==1.2.1
+    - cloudify-script-plugin
 
     \b
     Note:
-    - If source is URL, download and extract it and get module name and version
+    - If source is URL, download and extract it and get its name and version
      from setup.py.
-    - If source is a local path, get module name and version from setup.py.
-    - If source is module_name==module_version, use them as name and version.
+    - If source is a local path, get package name and version from setup.py.
+    - If source is `package_name==package_version`,
+     use them as name and version.
+    - If source is `package_name`, get the latest version from PyPI
+     (even if prerelease).
     """
     # TODO: Let the user provide supported Python versions.
     # TODO: Let the user provide supported Architectures.
+    logger.configure()
     packager = Wagon(source, verbose)
     packager.create(
         with_requirements, force, keep_wheels, exclude, output_directory,
-        pyver, validate)
+        pyver, validate, wheel_args, format)
 
 
 @click.command()
 @click.option('-s', '--source', required=True,
               help='Path or URL to source Wagon archive.')
-@click.option('--virtualenv', default=None,
+@click.option('-e', '--virtualenv', default=None,
               help='Virtualenv to install in.')
 @click.option('-r', '--requirements-file', required=False,
               help='A requirements file to install.')
 @click.option('-u', '--upgrade', required=False, is_flag=True,
-              help='Upgrades the module if it is already installed.')
+              help='Upgrades the package if it is already installed.')
 @click.option('--ignore-platform', required=False, is_flag=True,
               help='Ignores supported platform check.')
+@click.option('-a', '--install-args', required=False,
+              help='Allows to pass additional arguments to `pip install`. '
+                   '(e.g. -i my_pypi_index --retries 5')
 @click.option('-v', '--verbose', default=False, is_flag=True)
 def install(source, virtualenv, requirements_file, upgrade,
-            ignore_platform, verbose):
+            ignore_platform, install_args, verbose):
     """Installs a Wagon archive.
     """
+    logger.configure()
     installer = Wagon(source, verbose)
-    installer.install(virtualenv, requirements_file, upgrade, ignore_platform)
+    installer.install(
+        virtualenv, requirements_file, upgrade, ignore_platform, install_args)
 
 
 @click.command()
@@ -466,12 +526,13 @@ def install(source, virtualenv, requirements_file, upgrade,
 def validate(source, verbose):
     """Validates an archive.
 
-    This tests that all requires wheels exist, that the module
+    This tests that all requires wheels exist, that the package
     is installable and that different metadata properties exist.
     """
+    logger.configure()
     validator = Wagon(source, verbose)
     if not validator.validate():
-        sys.exit('validation_failed')
+        sys.exit(codes.errors['validation_failed'])
 
 
 @click.command()
@@ -481,6 +542,7 @@ def validate(source, verbose):
 def showmeta(source, verbose):
     """Prints out the metadata for an archive.
     """
+    logger.configure()
     getter = Wagon(source, verbose)
     metadata = getter.get_metadata_from_archive()
     print(json.dumps(metadata, indent=4, sort_keys=True))
